#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "config.h"
#include "emu.h"

#define DOS_GET_SDA_POINTER    0x5D06

#define Mah	1
#define Mal	2
#define Mbh	3
#define Mbl	4
#define Mch	5
#define Mcl	6
#define Mdh	7
#define Mdl	8

static FILE *out = 0;
static unsigned short * stack;
static struct vm86_regs saved_regs;
static int linux_fds[256];
static int lfn_seg;

/* Drives in this module are *always* A=1 B=2 etc */
static int lfndrive = -1;
static int current_drive = -2;

static void (*follow_up_func)(void) = 0;

static char lfn_cwd[1000];

typedef struct {
  char header[128];
  char buf[128];
  long jmp_start;
  short helper;
  short chain_ip;
  short chain_cs;
} tsr_s;
static tsr_s *tsr;

static char *
pathcomp(char *p, char *c)
{
  int l = strlen(c);
  if (memcmp(p, c, l) == 0
      && (p[l] == '/'
	  || p[l] == '\\'
	  || p[l] == 0))
    {
      while (p[l] == '/' || p[l] == '\\')
	l++;
      return p+l;
    }
  return 0;
}

static char *
backpath(char *p)
{
  while (*p == '\\' || *p == '/')
    p--;
  while (*p && *p != '\\' && *p != '/')
    p--;
  while (p[-1] == '/' || p[-1] == '\\')
    p--;
  if (!*p)
    *++p = '/';
  return p+1;
}

static char *
merge_paths(char *base_path, char *rel_path)
{
  static char *buf = 0;
  static int buflen = 0;
  char *s, *d, *t;
  int len;

  len = strlen(base_path) + strlen(rel_path) + 2;
  if (buflen < len)
    {
      buflen = len+10;
      if (buf)
	buf = (char *)realloc(buf, len);
      else
	buf = (char *)malloc(len);
    }

  buf[0] = 0;
  if (!rel_path)
    strcpy(buf+1, base_path);
  else if (*rel_path == '/' || *rel_path == '\\')
    strcpy(buf+1, rel_path);
  else
    {
      strcpy(buf+1, base_path);
      strcat(buf+1, "/");
      strcat(buf+1, rel_path);
    }
  s = d = buf+1;

  while (*s == '\\' || *s == '/')
    s++;

  while (*s)
    {
      if ((t=pathcomp(s, ".")))
	s = t;
      else if ((t=pathcomp(s, "..")))
	{
	  d = backpath(d);
	  s = t;
	}
      else
	{
	  *d++ = '/';
	  while (*s && *s != '\\' && *s != '/')
	    *d++ = *s++;
	  while (*s == '\\' || *s == '/')
	    s++;
	}
    }
  if (d == buf+1)
    *d++ = '/';
  *d = 0;

  return buf+1;
}

#if 0
testpath(char *a, char *b)
{
  char *c;
  printf("\"%s\" + \"%s\" ->", a, b);
  fflush(stdout);
  c = merge_paths(a,b);
  printf(" \"%s\"\n", c);
}
#endif

void
lfnnet_plugin_init(void)
{
  int i;
  dbug_printf("lfnnet: init\n");
  out = fdopen(2, "w");
  fprintf(out, "lfnnet: init\n");

  for (i=0; i<256; i++)
    linux_fds[i] = -1;

#if 0
  testpath("/", ".");
  testpath("/", "..");
  testpath("/", "a");
  testpath("/", "a/b");
  testpath("/a", ".");
  testpath("/a", "..");
  testpath("/a", "a");
  testpath("/a", "a/b");
  testpath("//a/", ".");
  testpath("//a/", "..");
  testpath("//a/", "a");
  testpath("//a/", "a/b");
  exit(0);
#endif

  getcwd(lfn_cwd, sizeof(lfn_cwd));
}

static void
setup_iret(int adjust)
{
  _IP = stack[1];
  _CS = stack[2];
  LWORD(esp) += adjust;
}

static int
lfndrivespec(char *str)
{
  if (str[1] != ':')
    return current_drive == lfndrive;
  if (str[0] == 'A'-1 + lfndrive)
    return 1;
  if (str[0] == 'a'-1 + lfndrive)
    return 1;
  return 0;
}

#define LFNDRIVENO(x) (((x) == 0 && current_drive == lfndrive) \
	|| (x) == current_drive)

#define IRET() {setup_iret(8); return;}
#define CHAIN() return;
#define INT21(x) { follow_up_func = x; REG(ip) = tsr->helper; return; }

static void
set_current_drive(void)
{
  current_drive = LO(dx) + 1;
  if (current_drive == lfndrive)
    {
      LO(ax) = lfndrive - 1;
      IRET();
    }
}

static void
get_current_drive(void)
{
  if (current_drive == lfndrive)
    {
      LO(ax) = current_drive - 1;
      IRET();
    }
}

static void
lfn_reset_drive(void)
{
  LWORD(eax) = 0x0d00;
  CHAIN();
}

static void
lfn_change_directory(void)
{
  char *src = SEG_ADR((char *), ds, si);
  fprintf(out, "LFN: chdir `%s'\n", src);
  LWORD(eax) = 0x3b00;
  CHAIN();
}

static void
lfn_get_current_directory(void)
{
  printf("cur %d lfn %d dl %d\n", current_drive, lfndrive, LO(dx));
  if (LFNDRIVENO(LO(dx)))
    {
      char *dst = SEG_ADR((char *), ds, si);
      strncpy(dst, lfn_cwd, 260);
      fprintf(out, "cwd `%s'\n", lfn_cwd);
      NOCARRY;
      IRET();
    }
  LWORD(eax) = 0x4700;
  CHAIN();
}

static void
lfn_find_first(void)
{
  CARRY;
  IRET();
}

static void
lfn_find_next(void)
{
  CARRY;
  IRET();
}

static void
lfn_find_close(void)
{
  CARRY;
  IRET();
}

static void
lfn_true_name(void)
{
  char *src = SEG_ADR((char *), ds, si);
  char *dst = SEG_ADR((char *), es, di);
  fprintf(out, "LFN: truename `%s'\n", src);
  strcpy(dst, src);
  NOCARRY;
  IRET();
}

static void
lfn_create_open(void)
{
  char *src = SEG_ADR((char *), ds, si);
  fprintf(out, "LFN: create/open `%s'\n", src);
  CARRY;
  IRET();
}

static void
get_lfn_info(void)
{
  char *src = SEG_ADR((char *), ds, dx);
  char *dst = SEG_ADR((char *), es, di);
  fprintf(out, "LFN: get lfn info `%s'\n", src);
  NOCARRY;
  if (lfndrivespec(src))
    {
      LWORD(ebx) = 0x4003; /* lfn, case sensitive and preserving */
      LWORD(ecx) = 255;
      LWORD(edx) = 260;
      strcpy(dst, "LINUX");
    }
  else
    {
      LWORD(ebx) = 0x0000; /* sfn, case insensitive and ignoring */
      LWORD(ecx) = 13;
      LWORD(edx) = 64;
      strcpy(dst, "SFN");
    }
  IRET();
}

static void
lfn_short_name(void)
{
  char *src = SEG_ADR((char *), ds, si);
  char *dst = SEG_ADR((char *), es, di);
  fprintf(out, "LFN: get sfn `%s'\n", src);
  strcpy(dst, src);
  NOCARRY;
  IRET();
}

static void
lfn_fail(void)
{
  LWORD(eax) = 0x7100;
  CARRY;
  IRET();
}

static struct {
  void (*handler)(void);
  char m[16];
} handlers[] = {
  { set_current_drive,		{ Mah, 0x0e }},
  { get_current_drive,		{ Mah, 0x19 }},
  { lfn_reset_drive,		{ Mah, 0x71, Mal, 0x0d }},
  { lfn_change_directory,	{ Mah, 0x71, Mal, 0x3b }},
  { lfn_get_current_directory,	{ Mah, 0x71, Mal, 0x47 }},
  { lfn_find_first,		{ Mah, 0x71, Mal, 0x4e }},
  { lfn_find_next,		{ Mah, 0x71, Mal, 0x4f }},
  { lfn_find_close,		{ Mah, 0x71, Mal, 0xa1 }},
  { lfn_true_name,		{ Mah, 0x71, Mal, 0x60 }},
  { lfn_create_open,		{ Mah, 0x71, Mal, 0x6c }},
  { get_lfn_info,		{ Mah, 0x71, Mal, 0xa0 }},
  { lfn_short_name,		{ Mah, 0x71, Mal, 0xa8 }},
  { lfn_fail,			{ Mah, 0x71 }},
  { 0 }
};

int
lfnnet_plugin_inte6(void)
{
  int my_ah = HI(ax);
  int i, mi;
  void (*old_follow_up_func)(void);

  stack = SEG_ADR((unsigned short *), ss, sp);
  LWORD(eax) = stack[0];

  switch (my_ah)
    {
    case 0:
      lfndrive = LO(dx);
      fprintf(out, "lfnnet: using drive %c:\n", '@' + lfndrive);
      LO(dx) = lfndrive;
      lfn_seg = REG(cs);
      tsr = (tsr_s *)(lfn_seg * 16 + 0x0108);
      break;

    case 1:
      saved_regs = REGS;
      follow_up_func = 0;

      for (i=0; handlers[i].handler; i++)
	{
	  for (mi=0; ; mi += 2)
	    switch (handlers[i].m[mi])
	      {
	      case Mah: if (HI(ax) != handlers[i].m[mi+1]) goto nomatch; break;
	      case Mal: if (LO(ax) != handlers[i].m[mi+1]) goto nomatch; break;
	      case Mbh: if (HI(bx) != handlers[i].m[mi+1]) goto nomatch; break;
	      case Mbl: if (LO(bx) != handlers[i].m[mi+1]) goto nomatch; break;
	      case Mch: if (HI(cx) != handlers[i].m[mi+1]) goto nomatch; break;
	      case Mcl: if (LO(cx) != handlers[i].m[mi+1]) goto nomatch; break;
	      case Mdh: if (HI(dx) != handlers[i].m[mi+1]) goto nomatch; break;
	      case Mdl: if (LO(dx) != handlers[i].m[mi+1]) goto nomatch; break;
	      case 0:
		fprintf(out, "lfnnet: pre ax=%04x bx=%04x cx=%04x dx=%04x  %04x:%04x\n",
			LWORD(eax), LWORD(ebx), LWORD(ecx), LWORD(edx),
			stack[2], stack[1]);
		handlers[i].handler();
		return 1;
	      }
	nomatch:
	}
      return 1;

    case 2:
      if (!follow_up_func)
	return 1;
      fprintf(out, "            ax=%04x bx=%04x cx=%04x dx=%04x            %04x\n",
	      LWORD(eax), LWORD(ebx), LWORD(ecx), LWORD(edx), LWORD(eflags));
      old_follow_up_func = follow_up_func;
      follow_up_func = 0;
      old_follow_up_func();
      return 1;
    }

  return 1;
}
